# -*- coding: utf-8 -*-
"""
/***************************************************************************
 FeatureZSetter
                                 A QGIS plugin
 Sets the Z value of each new/edited feature to a specific value, based on a DEM layer plus an offset. 
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2019-02-05
        git sha              : $Format:%H$
        copyright            : (C) 2019 by Centro de Informação Geoespacial do Exército 
        email                : igeoe@igeoe.pt
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from PyQt5.QtCore import QSettings, QTranslator, qVersion, QCoreApplication, Qt, QObject
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QAction, QToolBar, QMessageBox 

from qgis.gui import QgsMessageBar
from qgis.core import *

import re
from osgeo import gdal

# Initialize Qt resources from file resources.py
from .resources import *

# Import the code for the DockWidget
from .Feature_Z_Setter_dockwidget import FeatureZSetterDockWidget
import os.path


class FeatureZSetter:
    """QGIS Plugin Implementation."""

    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        # Save reference to the QGIS interface
        self.iface = iface

        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)

        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'FeatureZSetter_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)

            if qVersion() > '4.3.3':
                QCoreApplication.installTranslator(self.translator)

        # Declare instance attributes
        self.actions = []
        self.menu = self.tr(u'&Feature Z Setter')
        # TODO: We are going to let the user set this up in a future iteration

        # Check for CIGeoE toolbar. If exists, add button there; if not exists, create one
        cigeoeToolBarExists = False
        for x in iface.mainWindow().findChildren(QToolBar): 
            if x.windowTitle() == 'CIGeoE':
                self.toolbar = x
                cigeoeToolBarExists = True
        if cigeoeToolBarExists==False:
            self.toolbar = self.iface.addToolBar(u'CIGeoE')
        
        self.toolbar.setObjectName(u'FeatureZSetter')

        #print "** INITIALIZING FeatureZSetter"

        self.pluginIsActive = False
        self.dockwidget = None


        # my instance variables
        self.demLayerName = NULL
        self.demGDALLayer = NULL
        self.demLayerGeoTransformed = NULL
        self.layerBeingEdited = NULL
        self.noDEMLayerName = "** None **"



    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('FeatureZSetter', message)


    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            self.toolbar.addAction(action)

        if add_to_menu:
            self.iface.addPluginToMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action


    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        icon_path = ':/plugins/feature_z_setter/icon.png'
        self.add_action(
            icon_path,
            text=self.tr(u'Feature Z Setter: Set the Z value of features based on a DEM '),
            callback=self.run,
            parent=self.iface.mainWindow())

    #--------------------------------------------------------------------------

    def onClosePlugin(self):
        """Cleanup necessary items here when plugin dockwidget is closed"""

        #print "** CLOSING FeatureZSetter"

        # disconnects
        ##self.dockwidget.closingPlugin.disconnect(self.onClosePlugin)  
        ##self.dockwidget.demLayerCombo.currentIndexChanged.disconnect(self.newDEMLayerSelected)
        ##QObject.disconnect(self.iface, SIGNAL("currentLayerChanged(QgsMapLayer *)"), self.setLayerBeingEditedVar)
        ##if self.layerBeingEdited is not None:		
        ##    self.layerBeingEdited.featureAdded.disconnect( self.featureAddedTriggered )
        ##    self.layerBeingEdited.committedFeaturesAdded.disconnect( self.committedFeaturesAddedTriggered )

        #self.dockwidget.closingPlugin.disconnect()                                             # alterado
        self.dockwidget.closingPlugin.disconnect(self.onClosePlugin)

        #self.dockwidget.demLayerCombo.currentIndexChanged.disconnect()
        self.reconnect(self.dockwidget.demLayerCombo.currentIndexChanged)
        #QObject.disconnect(self.iface, SIGNAL("currentLayerChanged(QgsMapLayer *)"), self.setLayerBeingEditedVar)     # pyqgis2
        self.iface.currentLayerChanged.connect(self.setLayerBeingEditedVar)

        if self.layerBeingEdited is not NULL:
            #self.layerBeingEdited.featureAdded.disconnect()
            self.reconnect(self.layerBeingEdited.featureAdded)
            #self.layerBeingEdited.committedFeaturesAdded.disconnect()
            #self.reconnect(self.layerBeingEdited.committedFeaturesAdded)


        # remove this statement if dockwidget is to remain
        # for reuse if plugin is reopened
        # Commented next statement since it causes QGIS crashe
        # when closing the docked window:
        # self.dockwidget = None

        self.pluginIsActive = False
        self.dockwidget = None

    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""

        #print "** UNLOAD FeatureZSetter"

        for action in self.actions:
            self.iface.removePluginMenu(
                self.tr(u'&Feature Z Setter'),
                action)
            self.iface.removeToolBarIcon(action)
        # remove the toolbar
        del self.toolbar

    #--------------------------------------------------------------------------

    def run(self):
        """Run method that loads and starts the plugin"""

        #*************************************************************

        layersx = QgsProject.instance().layerTreeRoot().layerOrder()     
        
        if not layersx:        
            QMessageBox.information(self.iface.mainWindow(), "Error", 'Layer is not loaded!')            
            return 

        #*************************************************************


        if not self.pluginIsActive:
            self.pluginIsActive = True

            #print "** STARTING FeatureZSetter"

            # dockwidget may not exist if:
            #    first run of plugin
            #    removed on close (see self.onClosePlugin method)
            if self.dockwidget == None:
                # Create the dockwidget (after translation) and keep reference
                self.dockwidget = FeatureZSetterDockWidget()
                self.dockwidget.refreshButton.clicked.connect(self.refreshLayers)
                self.dockwidget.flattenButton.clicked.connect(self.setAllFeatureZCoordinatesToItsMaxValue)
                self.dockwidget.drapeButton.clicked.connect(self.setAllFeatureZCoordinatesToDEMValue)

                #setup initial values
                self.refreshLayers()
                # dont forget to disconnect onClosePlugin
                #self.dockwidget.demLayerCombo.currentIndexChanged.connect(self.newDEMLayerSelected)
                self.reconnect(self.dockwidget.demLayerCombo.currentIndexChanged, self.newDEMLayerSelected)
                #QObject.connect(self.iface, SIGNAL("currentLayerChanged(QgsMapLayer *)"), self.setLayerBeingEditedVar)     # pyqgis2
                self.iface.currentLayerChanged.connect(self.setLayerBeingEditedVar)

                # PyQt_PyObject
                # QObject.connect(self.iface, SIGNAL("featureAdded (QgsFeatureId fid)"), self.featureAddedTriggered)


            # connect to provide cleanup on closing of dockwidget
            self.dockwidget.closingPlugin.connect(self.onClosePlugin)

            # show the dockwidget
            # TODO: fix to allow choice of dock location
            self.iface.addDockWidget(Qt.LeftDockWidgetArea, self.dockwidget)
            self.dockwidget.show()
        else:
            self.iface.removeDockWidget(self.dockwidget)
            self.onClosePlugin()

    def reconnect(self, signal, newhandler=None, oldhandler=None):
        while True:
            try:
                if oldhandler is not None:
                    signal.disconnect(oldhandler)
                else:
                    signal.disconnect()
            except TypeError:
                break
        if newhandler is not None:
            signal.connect(newhandler)

    def refreshLayers(self):
        if not self.pluginIsActive:
            self.run()
            return
        self.demLayerName = NULL
        self.loadLayersIntoComboBox()
        self.setLayerBeingEditedVar()
        return

    def loadLayersIntoComboBox(self):
        #layers = self.iface.legendInterface().layers()                 # pyqgis2
        layers = QgsProject.instance().layerTreeRoot().layerOrder()
        layers_list=[]
        layers_list.append(self.noDEMLayerName)
        for layer in layers:
            if layer.type() == QgsMapLayer.RasterLayer:
                layers_list.append(layer.name())
            # DEBUG: self.iface.messageBar().pushMessage("Error", str(layer.type()), level=QgsMessageBar.CRITICAL)
        self.dockwidget.demLayerCombo.clear()
        self.dockwidget.demLayerCombo.addItems(layers_list)
        # set new dem selected layer
        self.newDEMLayerSelected()
        return

    def newDEMLayerSelected(self):
        # add field distinct values to values combo box
        selectedLayerIndex = self.dockwidget.demLayerCombo.currentIndex()
        if(selectedLayerIndex<0):
            return
        self.demLayerName=self.dockwidget.demLayerCombo.currentText()
        layerChecked = False
        #checks which is the selected layer. Only a raster layer can be a demLayer
        #layers = self.iface.legendInterface().layers()                # pyqgis2  
        layers = QgsProject.instance().layerTreeRoot().layerOrder()
        for layer in layers:
            #if there's a match, check how many channels has the raster (a DEM just have one)
            if layer.name()==self.demLayerName:
                layerPath=layer.dataProvider().dataSourceUri()
                self.demGDALLayer = gdal.Open(layerPath)
                bands = self.demGDALLayer.RasterCount
                if bands==1:
                    #having just one layer means it's a DEM, so get its geo transform to be ready to get it's pixel values
                    self.demLayerGeoTransformed = self.demGDALLayer.GetGeoTransform()
                    layerChecked=True
                    QgsMessageLog.logMessage("Layer \""+layer.name()+"\" opened as a DEM, new Z will take it into consideration.", 'FeatureZSetter')
                    break
        if layerChecked==False:
            self.demLayerGeoTransformed = NULL
            self.demGDALLayer = NULL
        return

    def setLayerBeingEditedVar(self):
        tempLayer=self.iface.activeLayer()
        #if tempLayer is not None and self.iface.activeLayer().isEditable():
        if tempLayer is not None and tempLayer.type() == QgsMapLayer.VectorLayer:
            self.layerBeingEdited=tempLayer
            #self.layerBeingEdited.featureAdded.connect( self.featureAddedTriggered )
            
            #layer commented 04JAN19
            #self.reconnect(self.layerBeingEdited.featureAdded, self.featureAddedTriggered)
            
            #self.layerBeingEdited.committedFeaturesAdded.connect( self.committedFeaturesAddedTriggered )
            #self.reconnect(self.layerBeingEdited.committedFeaturesAdded, self.committedFeaturesAddedTriggered)
            # eventualmente apagar esta message...
            if self.iface.activeLayer().isEditable():
                QgsMessageLog.logMessage(self.layerBeingEdited.name()+" is being tracked for changes", 'FeatureZSetter')
                #self.iface.messageBar().pushMessage("Feature Z Setter", self.layerBeingEdited.name()+" is being tracked for changes", level=QgsMessageBar.INFO)
        else:
            self.layerBeingEdited=NULL
        return

    # not used 04JAN19
    def featureAddedTriggered(self, fid):
        #must only accept geometries with Z; geom.wkbType() must be one of the following values:
        #PointZ = 1001, LineStringZ = 1002, PolygonZ = 1003, PointZM = 3001, LineStringZM = 3002, PolygonZM = 3003
        feat = NULL
        for f in self.layerBeingEdited.getFeatures():
            if f.id()==fid:
                feat=f
                break
        if feat is not None:
		    # get feature geometry
            geom = feat.geometry()
            #prevent invalid geometries
            #geomTypeInWkt=geom.exportToWkt().split('(')[0]                      # pyqgis2
            geomTypeInWkt=geom.asWkt().split('(')[0]
            if self.isAValidGeometryForThisPlugin(geomTypeInWkt) is False:
                QgsMessageLog.logMessage( "User tried to add a geometry that's not a point, linestring or polygon, or that does not have a Z coordinate (geometry: "+geomTypeInWkt+").", 'FeatureZSetter')
                return

            #QgsMessageLog.logMessage( "User added geometry: "+geom.exportToWkt(), 'FeatureZSetter')       # pyqgis2 
            QgsMessageLog.logMessage( "User added geometry: "+geom.asWkt(), 'FeatureZSetter')

            #will save Polygon, Point, LineString; the 3 with Z; the 3 with ZM
            #wktPrefix=geom.exportToWkt().split('(')[0].strip()                # pyqgis2
            wktPrefix=geom.asWkt().split('(')[0].strip()
            numberOfCoords=2
            if wktPrefix[-2:].lower()=='zm':
                numberOfCoords=4
            if wktPrefix[-1:].lower()=='z':
                numberOfCoords=3
            #QgsMessageLog.logMessage( wktPrefix, 'FeatureZSetter')
            if numberOfCoords>2:
                #adicionada linha 04JAN19
                #wktFull=geom.exportToWkt()                  # pyqgis2
                wktFull=geom.asWkt()

                #wktAldrabado=geom.exportToWkt()+",lixo"                       # pyqgis2
                wktAldrabado=geom.asWkt()+",lixo"
                nums = re.findall(r'\d+(?:\.\d*)?', wktAldrabado.rpartition(',')[0])


                #to find out, from wkt, if is a polygon with holes
                polygonEnd=[]
                #- remove all spaces
                wktFullXpto=wktFull.replace(" ", "")
                #- count number of "),(", meaning end of a ring and start of another
                numRings=wktFullXpto.count("),(")
                ringIndexStart=0
                for numRingsIndex in range(numRings):
                    #- count number of commas of each ring
                    ringIndexEnd=wktFull.find("),(")
                    ringCommaCount=wktFull[ringIndexStart:ringIndexEnd].count(",")
                    ringIndexStart=ringIndexEnd+1
                    polygonEnd.append(ringCommaCount)


                allCoords=[]
                for ff in nums:
                    allCoords.append( ff )
                #coords = zip(*[iter(nums)] * 3)
                #allCoords=map(float, nums)
                #for c in allCoords:
                #    QgsMessageLog.logMessage( str(c), 'FeatureZSetter')
                
                #get offset value
                try:
                    offset=float(self.dockwidget.offsetValue.text())
                except ValueError:
                    offset=0
                    self.dockwidget.offsetValue.setText("0")
                    QgsMessageLog.logMessage( "FeatureZSetter: invalid offset value. Set to 0.", 'FeatureZSetter')
                #iteration starts in the first Z position, which is at index 2 (x, y, z)
                ##for i in range(2, len(allCoords), numberOfCoords):
                ##    allCoords[i]=offset
                    #QgsMessageLog.logMessage( str( allCoords[i] ), 'FeatureZSetter')

                #iteration by all coordinates: gets X and Y to check it's DEM value; then adds it, plus the offset, in the Z value
                xCoords=[]
                yCoords=[]
                #grabs/orders all X coords
                for i in range(0,len(allCoords), numberOfCoords):
                    xCoords.append(allCoords[i])
                #grabs/orders all Y coords
                for i in range(1,len(allCoords), numberOfCoords):
                    yCoords.append(allCoords[i])
                #sets properly all Z coords
                j=0
                for i in range(2,len(allCoords), numberOfCoords):
                    #get DEM value and add it, plus the offset, to Z coordinate
                    if self.demGDALLayer is not NULL and self.demLayerGeoTransformed is not NULL:
                        value=self.getZValueFromRaster(xCoords[j], yCoords[j], self.demGDALLayer, self.demLayerGeoTransformed)
                    else:
                        value=0
                    allCoords[i]=value+offset
                    j=j+1
                #eventually sets M value to 0
                if numberOfCoords==4:
                    for i in range(3,len(allCoords), numberOfCoords):
                        allCoords[i]=0
                newWkt=wktPrefix+"("
                isPolygon=False
                if len(wktPrefix)>=7 and wktPrefix[:7].lower()=='polygon':
                    isPolygon=True
                if isPolygon:
                    newWkt=newWkt+"("
                for x in range(len(allCoords)):
                    if x!=0:
                        newWkt=newWkt+" "
                    newWkt=newWkt+str(allCoords[x])
                    if x%numberOfCoords==numberOfCoords-1 and x!=len(allCoords)-1:
                        #check if it's end of polygon inner ring
                        if x/numberOfCoords in polygonEnd:
                            newWkt=newWkt+"),("
                        else:
                            newWkt=newWkt+","
                newWkt=newWkt+")"
                if isPolygon:
                    newWkt=newWkt+")"
                QgsMessageLog.logMessage( "FeatureZSetter updated to geometry: "+newWkt, 'FeatureZSetter')
                newGeom=QgsGeometry.fromWkt(newWkt)
                self.layerBeingEdited.changeGeometry(f.id(), newGeom)
            else:
                QgsMessageLog.logMessage( "FeatureZSetter: feature just has X and Y coordinates, no updates made.", 'FeatureZSetter')
        return

    def setAllFeatureZCoordinatesToDEMValue(self):
        tempLayer=self.iface.activeLayer()
        if tempLayer is not None and tempLayer.isEditable():
            features = tempLayer.selectedFeatures()
            if not features:
                QMessageBox.information(self.iface.mainWindow(), "Error", "There's no features selected in active layer.")
            else:
                changedGeometries=0
                noZValue = False
                for f in features:
                    geom = f.geometry()
                    #prevent invalid geometries
                    #geomTypeInWkt=geom.exportToWkt().split('(')[0]                     # pyqgis2 
                    geomTypeInWkt=geom.asWkt().split('(')[0]
                    if self.isAValidGeometryForThisPlugin(geomTypeInWkt) is False:
                        QgsMessageLog.logMessage( "User tried to drape a geometry that's not a point, linestring or polygon, or that does not have a Z coordinate (geometry: "+geomTypeInWkt+").", 'FeatureZSetter')
                        continue
                    #QgsMessageLog.logMessage( "Draping geometry: "+geom.exportToWkt(), 'FeatureZSetter')            # pyqgis2 
                    QgsMessageLog.logMessage( "Draping geometry: "+geom.asWkt(), 'FeatureZSetter')
                    #will save Polygon, Point, LineString; the 3 with Z; the 3 with ZM
                    #wktPrefix=geom.exportToWkt().split('(')[0].strip()                       # pyqgis2 
                    wktPrefix=geom.asWkt().split('(')[0].strip()
                    numberOfCoords=2
                    if wktPrefix[-2:].lower()=='zm':
                        numberOfCoords=4
                    if wktPrefix[-1:].lower()=='z':
                        numberOfCoords=3
                    if numberOfCoords>2:
                        #wktFull=geom.exportToWkt()                                               # pyqgis2 
                        wktFull=geom.asWkt()   
                        wktAldrabado=wktFull+",lixo"
                        nums = re.findall(r'\d+(?:\.\d*)?', wktAldrabado.rpartition(',')[0])
                        #to find out, from wkt, if is a polygon with holes
                        polygonEnd=[]
                        #- remove all spaces
                        wktFullXpto=wktFull.replace(" ", "")
                        #- count number of "),(", meaning end of a ring and start of another
                        numRings=wktFullXpto.count("),(")
                        #ringIndexStart=0
                        ringIndexStart=wktFull.find("),(")+1
                        for numRingsIndex in range(numRings):
                            #- count number of commas of each ring
                            ringIndexEnd=wktFull.find("),(", ringIndexStart)
                            ringCommaCount=wktFull[ringIndexStart:ringIndexEnd].count(",")
                            if numRingsIndex==0:
                                exteriorRingCommaCount=wktFull[0:ringIndexStart].count(",")
                                polygonEnd.append(exteriorRingCommaCount)
                                polygonEnd.append(exteriorRingCommaCount+ringCommaCount)
                            else:
                                polygonEnd.append(ringCommaCount+polygonEnd[numRingsIndex])
                            ringIndexStart=ringIndexEnd+1
                        allCoords=[]
                        for ff in nums:
                            allCoords.append( ff )
                        #get offset value
                        try:
                            offset=float(self.dockwidget.offsetValue.text())
                        except ValueError:
                            offset=0
                            self.dockwidget.offsetValue.setText("0")
                            QgsMessageLog.logMessage( "FeatureZSetter: invalid offset value. Set to 0.", 'FeatureZSetter')
                        #iteration by all coordinates: gets X and Y to check highest Z value; then sets it to all vertices
                        xCoords=[]
                        yCoords=[]
                        #grabs/orders all X coords
                        for i in range(0,len(allCoords), numberOfCoords):
                            xCoords.append(allCoords[i])
                        #grabs/orders all Y coords
                        for i in range(1,len(allCoords), numberOfCoords):
                            yCoords.append(allCoords[i])
                        #sets properly all Z coords
                        j=0
                        for i in range(2,len(allCoords), numberOfCoords):
                            #get DEM value and add it, plus the offset, to Z coordinate
                            if self.demGDALLayer is not NULL and self.demLayerGeoTransformed is not NULL:
                                value=self.getZValueFromRaster(xCoords[j], yCoords[j], self.demGDALLayer, self.demLayerGeoTransformed)
                            else:
                                value=0
                            allCoords[i]=value+offset
                            j=j+1
                        #eventually sets M value to 0
                        if numberOfCoords==4:
                            for i in range(3,len(allCoords), numberOfCoords):
                                allCoords[i]=0
                        newWkt=wktPrefix+"("
                        isPolygon=False
                        if len(wktPrefix)>=7 and wktPrefix[:7].lower()=='polygon':
                            isPolygon=True
                        if isPolygon:
                            newWkt=newWkt+"("
                        for x in range(len(allCoords)):
                            if x!=0:
                                newWkt=newWkt+" "
                            newWkt=newWkt+str(allCoords[x])
                            if x%numberOfCoords==numberOfCoords-1 and x!=len(allCoords)-1:
                                #check if it's end of polygon inner ring
                                if x/numberOfCoords in polygonEnd:
                                    newWkt=newWkt+"),("
                                else:
                                    newWkt=newWkt+","
                        newWkt=newWkt+")"
                        if isPolygon:
                            newWkt=newWkt+")"
                        QgsMessageLog.logMessage( "FeatureZSetter updated to geometry: "+newWkt, 'FeatureZSetter')
                        newGeom=QgsGeometry.fromWkt(newWkt)
                        tempLayer.changeGeometry(f.id(), newGeom)
                        changedGeometries=changedGeometries+1
                    else:
                        QgsMessageLog.logMessage( "FeatureZSetter: feature just has X and Y coordinates, no updates made.", 'FeatureZSetter')
                        noZValue = True
                if noZValue == True:
                    QMessageBox.information(self.iface.mainWindow(), "Error", "Feature(s) just have X and Y coordinates, no updates made.")
                else:
                    #self.iface.messageBar().pushMessage("Feature Z Setter", str(changedGeometries)+" geometries drapped.", level=QgsMessageBar.SUCCESS)     # pyqgis2
                    self.iface.messageBar().pushMessage("Feature Z Setter", str(changedGeometries)+" geometries drapped.", level=Qgis.Success)
                    # to work well with feature nodes Z tag, force the unselect and select of features, to refresh canvas
                    # z values
                    #tempLayer.setSelectedFeatures([])                              # pyqgis2
                    #tempLayer.setSelectedFeatures([k.id() for k in features])      # pyqgis2
                    tempLayer.selectByIds([])                           # tempLayer.select([])
                    tempLayer.selectByIds([k.id() for k in features])   # tempLayer.select([k.id() for k in features])
                    
        else:
            QMessageBox.information(self.iface.mainWindow(), "Error", "There's no layer selected or selected layer not editable.")
        return

    def setAllFeatureZCoordinatesToItsMaxValue(self):
        tempLayer=self.iface.activeLayer()
        if tempLayer is not None and tempLayer.isEditable():
            features = tempLayer.selectedFeatures()
            if not features:
                QMessageBox.information(self.iface.mainWindow(), "Error", "There's no features selected in active layer.")
            else:
                changedGeometries=0
                noZValue = False
                for f in features:
                    geom = f.geometry()
                    #prevent invalid geometries
                    #geomTypeInWkt=geom.exportToWkt().split('(')[0]                           # pyqgis2 
                    geomTypeInWkt=geom.asWkt().split('(')[0]   
                    if self.isAValidGeometryForThisPlugin(geomTypeInWkt) is False:
                        QgsMessageLog.logMessage( "User tried to flatten a geometry that's not a point, linestring or polygon, or that does not have a Z coordinate (geometry: "+geomTypeInWkt+").", 'FeatureZSetter')
                        continue			                
                    #QgsMessageLog.logMessage( "Flattening geometry: "+geom.exportToWkt(), 'FeatureZSetter')         # pyqgis2  
                    QgsMessageLog.logMessage( "Flattening geometry: "+geom.asWkt(), 'FeatureZSetter')                    
                    #will save Polygon, Point, LineString; the 3 with Z; the 3 with ZM
                    #wktPrefix=geom.exportToWkt().split('(')[0].strip()                           # pyqgis2 
                    wktPrefix=geom.asWkt().split('(')[0].strip() 
                    numberOfCoords=2
                    if wktPrefix[-2:].lower()=='zm':
                        numberOfCoords=4
                    if wktPrefix[-1:].lower()=='z':
                        numberOfCoords=3
                    if numberOfCoords>2:
                        #wktFull=geom.exportToWkt()                               # pyqgis2 
                        wktFull=geom.asWkt() 
                        wktAldrabado=wktFull+",lixo"
                        nums = re.findall(r'\d+(?:\.\d*)?', wktAldrabado.rpartition(',')[0])

                        #to find out, from wkt, if is a polygon with holes
                        polygonEnd=[]
                        #- remove all spaces
                        wktFullXpto=wktFull.replace(" ", "")
                        #- count number of "),(", meaning end of a ring and start of another
                        numRings=wktFullXpto.count("),(")
                        #ringIndexStart=0
                        ringIndexStart=wktFull.find("),(")+1
                        for numRingsIndex in range(numRings):
                            #- count number of commas of each ring
                            ringIndexEnd=wktFull.find("),(", ringIndexStart)
                            ringCommaCount=wktFull[ringIndexStart:ringIndexEnd].count(",")
                            if numRingsIndex==0:
                                exteriorRingCommaCount=wktFull[0:ringIndexStart].count(",")
                                polygonEnd.append(exteriorRingCommaCount)
                                polygonEnd.append(exteriorRingCommaCount+ringCommaCount)
                            else:
                                polygonEnd.append(ringCommaCount+polygonEnd[numRingsIndex])
                            ringIndexStart=ringIndexEnd+1
                        allCoords=[]
                        for ff in nums:
                            allCoords.append( ff )
                        #iteration by all coordinates: gets X and Y to check highest Z value; then sets it to all vertices
                        xCoords=[]
                        yCoords=[]
                        #grabs/orders all X coords
                        for i in range(0,len(allCoords), numberOfCoords):
                            xCoords.append(allCoords[i])
                        #grabs/orders all Y coords
                        for i in range(1,len(allCoords), numberOfCoords):
                            yCoords.append(allCoords[i])
                        zMax=0.0
                        #checks highest Z coord
                        for i in range(2,len(allCoords), numberOfCoords):
                            if i==2:
                                zMax=allCoords[i]
                            if float(allCoords[i])>float(zMax):
                                zMax=allCoords[i]
                        #sets highest Z coord
                        for i in range(2,len(allCoords), numberOfCoords):
                            allCoords[i]=zMax
                        #eventually sets M value to 0
                        if numberOfCoords==4:
                            for i in range(3,len(allCoords), numberOfCoords):
                                allCoords[i]=0
                        newWkt=wktPrefix+"("
                        isPolygon=False
                        if len(wktPrefix)>=7 and wktPrefix[:7].lower()=='polygon':
                            isPolygon=True
                        if isPolygon:
                            newWkt=newWkt+"("
                        for x in range(len(allCoords)):
                            if x!=0:
                                newWkt=newWkt+" "
                            newWkt=newWkt+str(allCoords[x])
                            if x%numberOfCoords==numberOfCoords-1 and x!=len(allCoords)-1:
                                #check if it's end of polygon inner ring
                                if x/numberOfCoords in polygonEnd:
                                    newWkt=newWkt+"),("
                                else:
                                    newWkt=newWkt+","
                        newWkt=newWkt+")"
                        if isPolygon:
                            newWkt=newWkt+")"
                        QgsMessageLog.logMessage( "FeatureZSetter updated to geometry: "+newWkt, 'FeatureZSetter')
                        newGeom=QgsGeometry.fromWkt(newWkt)
                        tempLayer.changeGeometry(f.id(), newGeom)
                        changedGeometries=changedGeometries+1
                    else:
                        QgsMessageLog.logMessage( "FeatureZSetter: feature just has X and Y coordinates, no updates made.", 'FeatureZSetter')
                        noZValue = True
                if noZValue == True:
                    QMessageBox.information(self.iface.mainWindow(), "Error", "Feature(s) just have X and Y coordinates, no updates made.")
                else:
                    #self.iface.messageBar().pushMessage("Feature Z Setter", str(changedGeometries)+" geometries flattened.", level=QgsMessageBar.SUCCESS)    # pyqgis2
                    self.iface.messageBar().pushMessage("Feature Z Setter", str(changedGeometries)+" geometries flattened.", level=Qgis.Success)
                    # to work well with feature nodes Z tag, force the unselect and select of features, to refresh canvas
                    # z values
                    #tempLayer.setSelectedFeatures([])                              # pyqgis2
                    #tempLayer.setSelectedFeatures([k.id() for k in features])      # pyqgis2
                    tempLayer.selectByIds([])                           # tempLayer.select([])
                    tempLayer.selectByIds([k.id() for k in features])   # tempLayer.select([k.id() for k in features])
                    
        else:
            QMessageBox.information(self.iface.mainWindow(), "Error", "There's no layer selected or selected layer not editable.")
        return

    def getZValueFromRaster(self, x, y, layer, layerGeoTransform):
        px = int((float(x) - float(layerGeoTransform[0])) / float(layerGeoTransform[1]))
        py = int((float(y) - float(layerGeoTransform[3])) / float(layerGeoTransform[5]))
        band = layer.GetRasterBand(1)
        data = band.ReadAsArray(px,py, 1, 1)
        try:
            res=data[0][0]
        except TypeError:
            res=0
            QgsMessageLog.logMessage( "FeatureZSetter: no DEM value available, just considering offset.", 'FeatureZSetter')
        return res

    def isAValidGeometryForThisPlugin(self, geomWkt):
        #cleanWkt=geomWkt.strip() => only strips whitespaces to the left or right
        cleanWkt=re.sub('[\s+]', '', geomWkt).lower() #here strips any whitespaces in the wkt
        if cleanWkt=="pointz" or cleanWkt=="pointzm" or cleanWkt=="linestringz" or cleanWkt=="linestringzm" or cleanWkt=="polygonz" or cleanWkt=="polygonzm":
            return True
        else:
            return False
            